from .interfaces import ICommand

class MissingHandler(Exception):
    def __init__(self, message: str):
        super().__init__(message)

class OverwritingHandler(Exception):
    def __init__(self, message: str):
        super().__init__(message)


__version__ = '0.2.1'


from .interfaces import IApos
from .interfaces import IEvent
from .interfaces import ICommand
from .interfaces import IQuery
from .interfaces import IResponse
from .interfaces import IEventHandler
from .interfaces import ICommandHandler
from .interfaces import IQueryHandler

from .core import Apos

from .exceptions import MissingHandler
from .exceptions import OverwritingHandler


IApos
IEvent
ICommand
IQuery
IResponse
IEventHandler
ICommandHandler
IQueryHandler
Apos
MissingHandler
OverwritingHandler

apos = Apos()



from typing import Type, List
from abc import ABC, abstractmethod
from dataclasses import dataclass


class IApos(ABC):

    @abstractmethod
    def publish_command(
        self,
        command: "ICommand"
    ) -> List["IEvent"]:
        pass

    @abstractmethod
    def subscribe_command(
        self,
        command_cls: Type["ICommand"],
        command_handler: "ICommandHandler"
    ):
        pass

    @abstractmethod
    def publish_event(
        self,
        event: "IEvent"
    ):
        pass

    @abstractmethod
    def subscribe_event(
        self,
        event_cls: Type["IEvent"],
        event_handler: List["IEventHandler"]
    ):
        pass

    @abstractmethod
    def publish_query(
        self,
        query: "IQuery"
    ) -> "IResponse":
        pass

    @abstractmethod
    def subscribe_query(
        self,
        query_cls: Type["IQuery"],
        query_handler: "IQueryHandler"
    ):
        pass


@dataclass
class IEvent(ABC):
    pass


@dataclass
class ICommand(ABC):
    pass


@dataclass
class IQuery(ABC):
    pass


@dataclass
class IResponse(ABC):
    pass


class IEventHandler(ABC):
    def __call__(self, event: IEvent) -> None:
        pass


class ICommandHandler(ABC):
    def __call__(self, command: ICommand) -> None:
        pass


class IQueryHandler(ABC):
    def __call__(self, query: IQuery) -> IResponse:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apos']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'apos',
    'version': '0.2.1',
    'description': 'The backbone for message-driven applications.',
    'long_description': '# apos\n\nThe backbone for message-driven applications.\n\n## Summary\nThis Python library is designed to act as a software-level message broker, enabling a lightweight implementation of the publish-subscribe design pattern.\n\napos was born to accomplish two objectives:\n* Decouple the application layer from any interfaces\n* Develop reactive business functions\n\nWith apos, you can develop a message-driven application. This means that commands, events, and queries are sent to apos, which in return executes the functions that subscribe to these messages. This means that an adapter providing an external interface, may it be a web-API or a CLI, would not directly call application functions, but would rather send a message to apos, which will in return execute the business functions that subscribe to these messages. Equally, a business function would not call any other business function, but rather publishes an event, which other business functions can subscribe to and execute upon, controlled through apos.\n\n![](https://miro.medium.com/max/652/1*ZTxTLfH0FWRIQLAZFlBGEQ.png)\n\n## Context\nSee the Medium article linked below to read about why this library was created and how it is intended to be used. \nhttps://mkossatz.medium.com/a-backbone-for-message-driven-applications-ffdcef67824c\n\n\n## Installation\nThe library can be found on PyPi:\nhttps://pypi.org/project/apos/\n\n\n```shell\npip3 install apos\n```\n\n## Getting Started\n\nThe code below is a very lightweight example of how you can use apos for commands, queries, and events. \n\n```python\n\nfrom apos import apos\n\nclass RegisterUserCommand:\n    pass\n\nclass UserRegisteredEvent:\n    pass\n\nclass NewUserGreetedEvent:\n    pass\n\nclass RetrieveUserQuery:\n    pass\n\nclass RetrieveUserResponse:\n    pass\n\n\ndef register_user(command: RegisterUserCommand) -> None:\n    # Implementation of user registration\n    apos.publish_event(\n        UserRegisteredEvent())\n\n\ndef greet_new_user(event: UserRegisteredEvent) -> None:\n    # Implementation of user greeting\n    apos.publish_event(\n        NewUserGreetedEvent())\n\ndef retrieve_user(query: RetrieveUserQuery) -> RetrieveUserResponse:\n    # Implementation of user retrieval\n    return RetrieveUserResponse()\n\n# subscribing to messages (application configuration)\napos.subscribe_command(RegisterUserCommand, register_user)\napos.subscribe_event(UserRegisteredEvent, greet_new_user)\n\n# some interface adapter\napos.publish_command(RegisterUserCommand("Max"))\nevents = apos.get_published_events()\nprint(events)\n\n```\n\n\n\n## Complete Examples\n\nYou can find examples in the examples directory of the projects repository.\nhttps://github.com/mkossatz/apos/tree/main/examples',
    'author': 'Max Kossatz',
    'author_email': 'max@kossatzonline.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mkossatz/apos',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Type, Dict, List
from contextlib import contextmanager
from copy import deepcopy

from .interfaces import IApos
from .interfaces import IEvent
from .interfaces import ICommand
from .interfaces import IQuery
from .interfaces import IResponse
from .interfaces import IEventHandler
from .interfaces import ICommandHandler
from .interfaces import IQueryHandler
from .exceptions import MissingHandler
from .exceptions import OverwritingHandler


class Apos(IApos):

    def __init__(self) -> None:
        self._command_handlers: Dict[str, ICommandHandler] = dict()
        self._event_handlers: Dict[str, List[IEventHandler]] = dict()
        self._query_handlers: Dict[str, IQueryHandler] = dict()
        self._published_events: List[IEvent] = list()

    @contextmanager
    def session(self):
        published_events_bkp = deepcopy(self._published_events)
        self._published_events = list()
        try:
            yield self
        finally:
            self._published_events = published_events_bkp

    def get_published_events(self) -> List[IEvent]:
        return self._published_events

    def flush_published_events(self) -> None:
        self._published_events = list()

    def publish_command(
        self,
        command: ICommand
    ) -> None:
        command_name: str = command.__class__.__name__
        if command_name not in self._command_handlers:
            raise MissingHandler(
                "Publishing command {} failed because of missing subscriber.".format(command))
        command_handler: ICommandHandler = self._command_handlers[command_name]
        command_handler(command)

    def subscribe_command(
        self,
        command_cls: Type[ICommand],
        command_handler: ICommandHandler
    ) -> None:
        command_name: str = command_cls.__name__
        if command_name in self._command_handlers:
            raise OverwritingHandler(
                "A handler for the command {} already exists.".format(command_name))
        self._command_handlers[command_name] = command_handler

    def publish_event(
        self,
        event: IEvent
    ) -> None:
        event_name: str = event.__class__.__name__
        self._published_events.append(event)
        event_handlers: List[IEventHandler] = \
            self._event_handlers.get(event_name, [])
        for event_handler in event_handlers:
            event_handler(event)

    def subscribe_event(
        self,
        event_cls: Type[IEvent],
        event_handlers: List[IEventHandler]
    ) -> None:
        event_name: str = event_cls.__name__
        if event_name not in self._event_handlers.keys():
            self._event_handlers[event_name] = []
        for event_handler in event_handlers:
            if event_handler not in self._event_handlers[event_name]:
                self._event_handlers[event_name].append(event_handler)

    def publish_query(
        self,
        query: IQuery
    ) -> IResponse:
        query_name: str = query.__class__.__name__
        if query_name not in self._query_handlers:
            raise MissingHandler(
                "Publishing query {} failed because of missing subscriber.".format(query))
        query_handler: IQueryHandler = self._query_handlers[query_name]
        response: IResponse = query_handler(query)
        return response

    def subscribe_query(
        self,
        query_cls: Type[IQuery],
        query_handler: IQueryHandler
    ) -> None:
        query_name: str = query_cls.__name__
        if query_name in self._query_handlers:
            raise OverwritingHandler(
                "A handler for the query {} already exists.".format(query_name))
        self._query_handlers[query_name] = query_handler
